#!/usr/bin/env perl


package lib;

use URI::Encode;
use URI;
use JSON;

$JSON = JSON->new;

sub getUrlOptions
{
    my $argument = shift;
    my $options = {
        'schemes' => ["*"],
        'write' => ["*"],
    };
    foreach my $flagDef (split(/:/, $argument)) {
        my ($flag, $flagVal) = split(/=/, $flagDef, 2);
        if ($flag eq 'schemes' or $flag eq 'write') {
            $flagVal = &getList($flagVal);
            return undef if not defined $flagVal;
            $options->{$flag} = $flagVal;
        } else {
            return undef;
        }
    }

    return $options;
}

sub getList
{
    my $listString = shift;
    return undef if (not defined $listString);
    return [] if ($listString eq '');
    my @list = split(',', $listString);
    return \@list;
}

sub inArray
{
    my $needle = shift;
    my $haystack = shift;
    foreach (@$haystack) {
        return 1 if $_ eq $needle;
    }
    return 0;
}

sub decodeQuery
{
    my $query = shift;
    my $arrayFields = shift;
    $arrayFields = [] if not defined $arrayFields;
    my $encoder = URI::Encode->new();
    my $data = {};
    if (defined $query) {
        foreach (split(/\&/, $query)) {
            my ($key, $value) = split(/=/, $_, 2);
            $key = $encoder->decode($key);
            $value = $encoder->decode($value) if defined $value;
            if (&inArray($key, $arrayFields)) {
                $data->{$key} = [] if not defined($data->{$key});
                push @{$data->{$key}}, $value;
            } else {
                $data->{$key} = $value;
            }
        }
    }
    return $data;
}


package url_schemes;

sub http
{
    return &_httpLike(shift, 80);
}

sub https
{
    return &_httpLike(shift, 443);
}

sub mysql
{
    my $vars = &_httpLike(shift, 3306);
    return undef if not defined $vars;
    return undef if(substr($vars->{path}, 0, 1) ne '/');
    my $database = substr($vars->{path}, 1);
    my $query = &lib::decodeQuery($vars->{query});
    return {
        "url" => $vars->{url},
        "host" => $vars->{host},
        "port" => $vars->{port},
        "username" => $vars->{username},
        "password" => $vars->{password},
        "database" => $database,
        "version" => $query->{serverVersion},
    }
}

sub rabbitmq
{
    my $vars = &_httpLike(shift, 5672);
    return undef if not defined $vars;
    return undef if(substr($vars->{path}, 0, 1) ne '/');
    my $vhost = substr($vars->{path}, 1);
    return undef if not defined $vars->{username};
    return undef if not defined $vars->{password};
    my $query = &lib::decodeQuery($vars->{query}, ["topic"]);
    my $topics = undef;
    my $topicsJson = undef;
    if (defined $query->{topic}) {
        $topics = join(" ", grep(!/ /, @{$query->{topic}}));
        $topicsJson = $lib::JSON->encode($query->{topic});
    }
    return {
        "url" => $vars->{url},
        "host" => $vars->{host},
        "port" => $vars->{port},
        "username" => $vars->{username},
        "password" => $vars->{password},
        "vhost" => $vhost,
        "exchange" => $query->{exchange},
        "queue" => $query->{queue},
        "topics" => $topics,
        "topics_json" => $topicsJson,
    }
}

sub _any
{
    my $uri = shift;
    my $parsed = &_httpLike($uri);
    return $parsed if defined $parsed;
    if ($uri->has_recognized_scheme) {
        return {"url" => $uri->as_string};
    }
    return undef;
}

sub _httpLike
{
    my $encoder = URI::Encode->new();
    my $uriOrig = shift;
    my $uri = $uriOrig->clone;
    my $defaultPort = shift;
    my $scheme = $uri->scheme;
    $uri->scheme("http");
    my $host = $uri->host;
    my $port1 = $uri->port;
    $uri->scheme("https");
    my $port2 = $uri->port;
    $uri->scheme("http");
    my $port = ($port1 eq $port2) ? $port1 : $defaultPort;
    my $path = $uri->path;
    my $query = $uri->query;
    my $hash = $uri->fragment;
    my $userinfo = $uri->userinfo;
    my $username = undef;
    my $password = undef;
    if (defined $userinfo) {
        ($username, $password) = split(/:/, $userinfo, 2);
    }
    return undef if not defined $host or not defined $path;
    return undef if $path ne '' and $path !~ /^\//;
    $path = $encoder->decode($path) if defined $path;
    $username = $encoder->decode($username) if defined $username;
    $password = $encoder->decode($password) if defined $password;
    my $parsed = {
        "url" => $uriOrig->as_string,
        "scheme" => $scheme,
        "host" => $host,
        "port" => $port,
        "path" => $path,
        "query" => $query,
        "username" => $username,
        "password" => $password,
    }
}

package checkers;

sub credentials
{
    my $context = shift;
    my $value = $context->value;
    my ($username, $password);
    if (not defined $value or $value eq '') {
        $context->error("credentials cannot be empty");
        return;
    }
    ($username, $password) = split(/:/, $value, 2);
    if (not defined $password or not defined $username) {
        $context->error("credentials does not contain username:password");
        return;
    }

    my $encoder = URI::Encode->new();
    $username = $encoder->decode($username);
    $password = $encoder->decode($password);
    my $variable = $context->variableName;
    $context->put($variable . "_USERNAME", $username);
    $context->put($variable . "_PASSWORD", $password);
}

sub url
{
    my $context = shift;
    my $putDerivedEnvVars = 1;
    my $value = $context->value;
    my $options = &lib::getUrlOptions($context->argument);

    if (not defined $options) {
        $context->error("invalid url options");
        return;
    }
    return if (!$context->invoke("exists", $value));

    my $uri = URI->new($context->value, "http");
    my $scheme = $uri->scheme;

    if (defined $scheme and $scheme ne '' and $scheme !~ /^_/) {
        my $schemeOk = 0;
        foreach (@{$options->{schemes}}) {
            if (($_ eq '*') or ($_ eq $scheme)) {
                $schemeOk = 1;
                break;
            }
        }
        if (!$schemeOk) {
            $context->error("url scheme not allowed");
            return;
        }
        my $parser = "_any";
        if ($url_schemes::{$scheme}) {
            $parser = $scheme;
        }
        $parser = $url_schemes::{$parser};
        $uri = &$parser($uri);
    } else {
        $context->error("invalid url scheme");
        return;
    }

    if (not defined $uri) {
        $context->error("invalid url");
    }

    my %writes = ();
    foreach (@{$options->{write}}) {
        $writes{$_} = 1;
    }

    my $variable = $context->variableName;
    foreach (keys %{$uri}) {
        if ($writes{$_} or $writes{"*"}) {
            my $writeVar = $variable;
            $writeVar .= "_" . uc($_) if ($_ ne 'url');
            my $value = $uri->{$_};
            $context->put($writeVar, $value);
        }
    }
}

sub delete
{
    my $context = shift;
    if (defined $context->value) {
        $context->put($context->variableName, undef);
    }
}

sub default
{
    my $context = shift;
    my $value = $context->value;
    if (not defined $value or $value eq '') {
        $context->put($context->variableName, $context->argument);
    }
}

sub ne
{
    my $context = shift;
    my $value = $context->value;
    if (defined $value and $value eq $context->argument) {
        $context->error("Value forbidden");
    }
}

sub enum
{
    my $context = shift;
    my $value = $context->value;
    if (not defined $value) {
        $context->error("Missing value");
    } else {
        foreach ($context->argument("all")) {
            return if $value eq $_;
        }
        $context->error("Invalid value: $value");
    }
}

sub port
{
    my $context = shift;
    $context->invoke("uint", $context->value, "1:65535");
}

sub uint
{
    my $context = shift;
    $context->invoke("int", $context->value, "0") and $context->invoke("int", $context->value, $context->argument);
}

sub bool
{
    my $context = shift;
    my $value = $context->value;
    if (not defined $value or $value eq '') {
        $context->error("Missing boolean value");
    } else {
        $value = lc($value);
        if ($value eq '0' or $value eq 'false' or $value eq 'no') {
            $value="0";
        } elsif ($value eq '1' or $value eq 'true' or $value eq 'yes') {
            $value="1";
        } else {
            $context->error("Invalid boolean value");
        }
        $context->put($context->variableName, $value) if $context->value ne $value;
    }

}

sub int
{
    my $context = shift;
    my $min = $context->argument(0);
    my $max = $context->argument(1);
    my $value = $context->value;
    if (!&_intOk($min, 1) || !&_intOk($max, 1)) {
        $context->error("Invalid integer range");
    } elsif (!&_intOk($value, 0)) {
        $context->error("Invalid integer: " . $context->variableName);
    } elsif (defined $min && $value < $min) {
        $context->error("Integer must be at least $min: " . $context->variableName);
    } elsif (defined $max && $value > $max) {
        $context->error("Integer must be at most $max: " . $context->variableName);
    }
}

sub _intOk
{
    my $allowEmpty = $_[1];
    if (not defined $_[0] or $_[0] eq '') {
        return 0 if not $allowEmpty;
        $_[0] = undef;
        return 1;
    }
    return 0 if $_[0] !~ /^-?[0-9]+$/;
    return 1;
}

sub re
{
    my $context = shift;
    my $re = $context->argument;
    my $value = $context->value;
    if (defined $value) {
        if ($value !~ /$re/) {
            $context->error("Variable does not match expected criteria: " . $context->variableName);
        }
    } else {
        $context->error("Variable does not exist: " . $context->variableName); 
    }
}

sub nonempty
{
    my $context = shift;
    my $value = $context->value;
    if (not defined $value) {
        $context->error("Variable does not exist: " . $context->variableName);
    } elsif ($value eq '') {
        $context->error("Variable is empty: " . $context->variableName);
    }
}

sub exists 
{
    my $context = shift;
    $context->error("Variable does not exist: " . $context->variableName) if (not defined $context->value);
}

sub any
{
}

sub identifier
{
    my $context = shift;
    my $value = $context->value;
    if (not defined $value) {
        $context->error("Variable does not exist: " . $context->variableName);
    } elsif ($value eq '') {
        $context->error("Variable is empty: " . $context->variableName);
    } elsif ($value !~ /^[a-zA-Z_][a-zA-Z0-9_]*$/) {
        $context->error("Invalid identifier: " . $context->variableName);
    }
}

package context;

sub new
{
    return bless {
            "vars" => {},
            "error" => 0,
            "errorMessages" => [],
            "variableName" => undef,
            "value" => undef,
            "argument" => undef,
            "parsedArguments" => undef,
    }, shift
}

sub _setVariable
{
    my $self = shift;
    my $variableName = shift;
    my $value = shift;
    my $argument = shift;
    $self->{variableName} = $variableName;
    $self->{value} = $value;
    $self->{argument} = $argument;
    $self->{parsedArguments} = undef;
    return $self
}

sub variableName
{
    my $self = shift;
    return $self->{variableName};

}

sub put
{
    my $self = shift;
    my $variable = shift;
    my $value = shift;
    $self->{vars}->{$variable} = $value;
    return $self;
}

sub env
{
    my $self = shift;
    my $variable = shift;
    return $self->{vars}->{$variable} if (exists $self->{vars}->{$variable});
    return $ENV{$variable};
}

sub error
{
    my $self = shift;
    my $message = shift;

    $self->{error}++;
    push @{$self->{errorMessages}}, $message if defined $message;
    return $self;
}

sub argument
{
    my $self = shift;
    my $number = shift;
    if (not defined $number or not defined $self->{argument}) {
        return $self->{argument};
    }

    if (not defined $self->{parsedArguments}) {
        my @arguments = split(/[=:]/, $self->{argument});
        $self->{parsedArguments} = \@arguments;
    }

    if ($number eq 'all') {
        return @{$self->{parsedArguments}};
    } elsif($number =~ /^[0-9]+$/) {
        return $self->{parsedArguments}->[$number];
    } else {
        return undef;
    }

}

sub value
{
    my $self = shift;
    return $self->{value};
}

sub vars
{
    my $self = shift;
    my $changedOnly = shift;

    my %vars = ();
    foreach my $key (keys %{$self->{vars}}) {
        my $value = $self->{vars}->{$key};
        if ($changedOnly) {
            if (not defined $ENV{$key} and not defined $value) {
                next;
            }
            if (defined $ENV{$key} and defined $value and $ENV{$key} eq $value) {
                next;
            }
        }
        $vars{$key} = $value;
    }
    return \%vars;
}

sub errors
{
    my $self = shift;
    return undef if $self->{error} == 0;
    return $self->{errorMessages};
}

sub invoke
{
    my $self = shift;
    my $checker = shift;
    my $value = shift;
    my $argument = shift;
    my $variableName = shift;
    my @tmp = ();
    @tmp = ($self->{variableName}, $self->{value}, $self->{argument}, $self->{parsedArguments});
    $self->{variableName} = $variableName if (defined $variableName);
    $self->{value} = $value;
    $self->{argument} = $argument;
    $self->{parsedArguments} = undef;
    my $errorCounter = $self->{error};
    if ($checker !~ /^_/ and defined $checkers::{$checker}) {
        my $sub = $checkers::{$checker};
        &$sub($self);
    } else {
        $self->error("Invalid checker: $checker");
    }
    ($self->{variableName}, $self->{value}, $self->{argument}, $self->{parsedArguments}) = @tmp;
    return ($self->{error} == $errorCounter);
}

package main;

use Data::Dumper;
use Getopt::Long;
use String::ShellQuote 'shell_quote';

my $options = {};
GetOptions($options, "quiet", "q", "no-exit", "X");

my $context = context->new;
my $lastEnv = undef;
foreach (@ARGV) {
    my @matches = /^([a-zA-Z_][a-zA-Z0-9_]*|\.)([=:](([a-zA-Z_][a-zA-Z0-9_]*)([=:](.*))?))?$/;
    if (@matches) {
        my $env = $matches[0];
        my $checker = $matches[3];
        my $argument = $matches[5];
        if ($env eq '.') {
            $env = $lastEnv;
        }
        $lastEnv = $env;
        my $value = $context->env($env);
        $checker = "nonempty" if $checker eq '';
        $argument = undef if $argument eq '';
        if (defined $env) {
            $context->invoke($checker, $value, $argument, $env);
        } else {
            $context->error("Dot env variable cannot be used first");
        }
    } else {
        $context->error("Invalid checker descriptor: $_");
    }
}

my $vars = $context->vars(1);
my $errors = $context->errors;

if (not ($options->{quiet} or $options->{q})) {
    if (defined $errors) {
        foreach (@{$errors}) {
            print STDERR "Error: $_\n";
        }
    }
}

if (not defined $errors) {
    foreach my $key (keys %{$vars}) {
        my $value = $vars->{$key};
        if (defined $value) {
            print "export $key=" . &shell_quote($value) . "\n";
        } else {
            print "unset $key\n";
        }
    }
} else {
    if (not ($options->{X} || $options->{"no-exit"})) {
        print "exit 1\n";
    }
}

exit((defined $errors) ? 1 : 0);

